from .agent import (
    Agent,
    AgentActivities,
    AgentCallParams,
    AgentExecuteFnResult,
    AgentInstructionOptions,
    AgentResponse,
    AgentOptions,
    AgentToolCall,
    ExecuteFunctionParams,
    ExecuteInstructionsParams,
    Interrupt,
    ResumeRequest,
    ResumeResponse,
)

try:
    from .openai_agent import OpenAIAgent, OpenAIAgentOptions, OpenAIAgentActivities  # noqa: F401

    _OPENAI_AVAILABLE_ = True
except ImportError:
    _OPENAI_AVAILABLE_ = False

try:
    from .anthropic_agent import (  # noqa: F401
        AnthropicAgent,
        AnthropicAgentOptions,
        AnthropicAgentActivities,
    )

    _ANTHROPIC_AVAILABLE_ = True
except ImportError:
    _ANTHROPIC_AVAILABLE_ = False

__all__ = [
    "Agent",
    "AgentActivities",
    "AgentOptions",
    "AgentCallParams",
    "AgentExecuteFnResult",
    "AgentInstructionOptions",
    "AgentResponse",
    "AgentToolCall",
    "ExecuteFunctionParams",
    "ExecuteInstructionsParams",
    "Interrupt",
    "ResumeRequest",
    "ResumeResponse",
]

if _OPENAI_AVAILABLE_:
    __all__.extend(
        [
            "OpenAIAgent",
            "OpenAIAgentOptions",
            "OpenAIAgentActivities",
        ]
    )

if _ANTHROPIC_AVAILABLE_:
    __all__.extend(
        [
            "AnthropicAgent",
            "AnthropicAgentOptions",
            "AnthropicAgentActivities",
        ]
    )



from dataclasses import dataclass, field
from typing import Any, Literal
from openai import NotGiven, OpenAI
from temporalio import activity
from rojak.utils import function_to_json, mcp_to_openai_tool
from rojak.agents import (
    Agent,
    AgentActivities,
    AgentCallParams,
    AgentExecuteFnResult,
    AgentResponse,
    AgentOptions,
    AgentToolCall,
    ExecuteFunctionParams,
    ExecuteInstructionsParams,
)
from openai.types.chat import ChatCompletion
import os


@dataclass
class OpenAIAgentOptions(AgentOptions):
    api_key: str | None = None
    client: OpenAI | None = None
    base_url: str | None = None
    inference_config: dict[str, Any] = field(
        default_factory=lambda: {
            "max_tokens": 1000,
            "temperature": 0.0,
            "top_p": 0.9,
            "stop_sequences": [],
        }
    )


@dataclass
class OpenAIAgent(Agent):
    model: str = "gpt-4o-mini"

    type: Literal["openai"] = field(default="openai")
    """Type of agent. Must be `"openai"`."""

    inference_config: dict[str, Any] | None = None
    """Inference configuration for OpenAI models"""


class OpenAIAgentActivities(AgentActivities):
    def __init__(self, options: OpenAIAgentOptions = OpenAIAgentOptions()):
        super().__init__(options)

        if options.client:
            self.client = options.client
        elif options.api_key:
            self.client = OpenAI(api_key=options.api_key, base_url=options.base_url)
        elif os.environ.get("OPENAI_API_KEY"):
            self.client = OpenAI(
                api_key=os.environ.get("OPENAI_API_KEY"), base_url=options.base_url
            )
        else:
            raise ValueError("OpenAI API key is required")
        self.inference_config = options.inference_config

    @staticmethod
    def handle_model_response(response: ChatCompletion) -> AgentResponse:
        """Convert model response to AgentResponse"""
        message = response.choices[0].message
        if message.tool_calls:
            tool_calls = [
                AgentToolCall(**dict(tool_call)) for tool_call in message.tool_calls
            ]
            return AgentResponse(
                content=message.content,
                tool_calls=tool_calls,
                type="tool",
            )
        elif message.content:
            return AgentResponse(content=message.content, type="text")
        else:
            raise ValueError("Unknown message type")

    @activity.defn(name="openai_call")
    async def call(self, params: AgentCallParams) -> AgentResponse:
        # Create list of messages
        messages = [
            {
                "role": msg.role,
                "content": msg.content,
                "tool_calls": msg.tool_calls,
                "tool_call_id": msg.tool_call_id,
            }
            for msg in params.messages
        ]

        # Update inference config if needed
        if params.inference_config:
            self.inference_config = {**self.inference_config, **params.inference_config}

        # Create tool call json
        functions = [
            self.function_map[name]
            for name in params.function_names
            if name in self.function_map
        ]
        tools = [function_to_json(f) for f in functions]
        for tool in tools:
            fn_params = tool["function"]["parameters"]
            fn_params["properties"].pop("context_variables", None)
            if "context_variables" in fn_params["required"]:
                fn_params["required"].remove("context_variables")

        tools += [mcp_to_openai_tool(tool) for tool in self.mcp_result.tools.values()]

        response = self.client.chat.completions.create(
            model=params.model,
            messages=messages,
            tools=tools or None,
            tool_choice=params.tool_choice,
            parallel_tool_calls=params.parallel_tool_calls if tools else NotGiven(),
            max_tokens=self.inference_config["max_tokens"],
            temperature=self.inference_config["temperature"],
            top_p=self.inference_config["top_p"],
            stop=self.inference_config["stop_sequences"],
        )

        return self.handle_model_response(response)

    @activity.defn(name="openai_execute_instructions")
    async def execute_instructions(self, params: ExecuteInstructionsParams) -> str:
        return await super().execute_instructions(params)

    @activity.defn(name="openai_execute_function")
    async def execute_function(
        self, params: ExecuteFunctionParams
    ) -> str | OpenAIAgent | AgentExecuteFnResult:
        return await super().execute_function(params)



from .orchestrator_workflow import (
    OrchestratorParams,
    OrchestratorResponse,
    UpdateConfigParams,
    OrchestratorWorkflow,
    GetConfigResponse,
    TaskParams,
)
from .agent_workflow import (
    AgentWorkflowRunParams,
    ToolResponse,
    AgentWorkflowResponse,
    AgentWorkflow,
    AgentTypes,
)

__all__ = [
    "OrchestratorParams",
    "OrchestratorResponse",
    "UpdateConfigParams",
    "OrchestratorWorkflow",
    "AgentWorkflowRunParams",
    "ToolResponse",
    "AgentWorkflowResponse",
    "AgentWorkflow",
    "AgentTypes",
    "GetConfigResponse",
    "TaskParams",
]



from dataclasses import dataclass
from datetime import timedelta
import json
from typing import TYPE_CHECKING, Union
from temporalio import workflow
from temporalio.exceptions import ActivityError
from rojak.retrievers import Retriever
from rojak.types import ContextVariables, ConversationMessage
from rojak.utils import create_retry_policy, debug_print
from rojak.agents import (
    AgentCallParams,
    ExecuteFunctionParams,
    ExecuteInstructionsParams,
    AgentInstructionOptions,
    AgentToolCall,
    AgentResponse,
    AgentExecuteFnResult,
    Agent,
    ResumeResponse,
    ResumeRequest,
)

if TYPE_CHECKING:
    from rojak.workflows.orchestrator_workflow import OrchestratorWorkflow

try:
    from rojak.agents import OpenAIAgent
except ImportError:
    OpenAIAgent = None

try:
    from rojak.agents import AnthropicAgent
except ImportError:
    AnthropicAgent = None

AgentTypes = (
    Union[*(agent for agent in (OpenAIAgent, AnthropicAgent) if agent is not None)]
    | Agent  # typing fallback
)


@dataclass
class AgentWorkflowRunParams:
    agent: AgentTypes
    """The agent to be called."""

    messages: list[ConversationMessage]
    """List of message objects."""

    context_variables: ContextVariables
    """A dictionary of additional context variables, available to functions and Agent instructions."""

    orchestrator: "OrchestratorWorkflow"
    """The parent orchestrator that called this workflow."""

    task_id: str
    """Unique identifier for this request."""

    debug: bool = False
    """If True, enables debug logging."""


@dataclass
class ToolResponse:
    tool_call_id: str
    """Unique identifier for the tool call the response is for."""

    output: AgentExecuteFnResult
    """Result from tool call function."""


@dataclass
class AgentWorkflowResponse:
    output: str | ToolResponse
    """Agent Workflow output."""

    sender: str
    """Indicate which agent the message originated from."""


class ToolRejectedError(Exception):
    """Error raised when a tool call is rejected."""

    def __init__(self, message: str, cause: Exception = None):
        super().__init__(message)
        self.__cause__ = cause


class AgentWorkflow:
    def __init__(self, params: AgentWorkflowRunParams):
        self.orchestrator = params.orchestrator
        self.task_id = params.task_id
        self.agent = params.agent
        self.retry_policy = create_retry_policy(self.agent.retry_options.retry_policy)
        self.start_to_close_timeout = timedelta(
            seconds=params.agent.retry_options.timeout_in_seconds
        )
        self.debug = params.debug
        self.messages = params.messages
        self.context_variables = params.context_variables

        # Handle interrupts
        self.interrupt_map = {
            interrupt.tool_name: interrupt for interrupt in params.agent.interrupts
        }
        self.interrupted: set[str] = set()  # tool call ids for approval
        self.resumed: dict[
            str, ResumeResponse
        ] = {}  # tool call ids that resumed, pending actions

    async def run(self) -> tuple[AgentWorkflowResponse, list[ConversationMessage]]:
        # process instructions
        instructions = self.agent.instructions
        try:
            if isinstance(instructions, AgentInstructionOptions):
                instructions: str = await workflow.execute_activity(
                    f"{self.agent.type}_execute_instructions",
                    ExecuteInstructionsParams(
                        instructions,
                        self.context_variables,
                    ),
                    result_type=str,
                    start_to_close_timeout=self.start_to_close_timeout,
                    retry_policy=self.retry_policy,
                )
        except ActivityError as e:
            workflow.logger.error(f"Failed to execute instructions: {e}")
            raise

        # augment instructions
        if isinstance(self.agent.retriever, Retriever):
            context_prompt = await self.retrieve_context(self.messages[-1])
            instructions += context_prompt

        # execute call model activity
        response: AgentResponse = await workflow.execute_activity(
            f"{self.agent.type}_call",
            AgentCallParams(
                messages=[
                    ConversationMessage(role="system", content=instructions),
                    *self.messages,
                ],
                model=self.agent.model,
                function_names=self.agent.functions,
                parallel_tool_calls=self.agent.parallel_tool_calls,
                tool_choice=self.agent.tool_choice,
            ),
            result_type=AgentResponse,
            start_to_close_timeout=self.start_to_close_timeout,
            retry_policy=self.retry_policy,
        )

        # dont use isinstance to check as response output type different for different llm providers
        if response.type == "tool":
            tool_calls = response.tool_calls
            self.messages.append(
                ConversationMessage(
                    role="assistant",
                    content=response.content,
                    tool_calls=tool_calls,
                    sender=self.agent.name,
                ),
            )
            debug_print(self.debug, workflow.now(), f"{self.agent.name}: {tool_calls}")

            # TODO: Figure out how to handle concurrent tool calls without race conditions in context_variables
            results: list[AgentWorkflowResponse] = []
            for tool_call in tool_calls:
                result = await self.handle_tool_call(tool_call, self.context_variables)
                assert isinstance(result.output, ToolResponse)
                self.context_variables = result.output.output.context_variables
                results.append(result)

            final_result: AgentWorkflowResponse | None = None
            for result in results:
                assert isinstance(result.output, ToolResponse)
                fn_result = result.output.output
                debug_print(
                    self.debug,
                    workflow.now(),
                    f"{self.agent.name}: {fn_result.output}",
                )
                self.messages.append(
                    ConversationMessage(
                        role="tool",
                        content=fn_result.output,
                        sender=self.agent.name,
                        tool_call_id=result.output.tool_call_id,
                    )
                )
                # Send the last tool call response back to orchestrator to be used for next call to agent
                # Check if any tool call response is an agent. If so, send it back to orchestrator for next call to agent
                # If there are multiple tool call returning an agent, the last one will be used.
                if fn_result.agent:
                    final_result = result

            if not final_result:
                final_result = results[-1]

        else:
            assert isinstance(response.content, str)
            self.messages.append(
                ConversationMessage(
                    role="assistant",
                    content=response.content,
                    sender=self.agent.name,
                )
            )
            debug_print(
                self.debug, workflow.now(), f"{self.agent.name}: {response.content}"
            )
            final_result = AgentWorkflowResponse(
                output=response.content, sender=self.agent.name
            )

        return final_result, self.messages

    async def retrieve_context(self, message: ConversationMessage) -> str:
        try:
            retriever_result: str = await workflow.execute_activity(
                f"{self.agent.retriever.type}_retrieve_and_combine_results",
                message.content,
                result_type=str,
                start_to_close_timeout=self.start_to_close_timeout,
                retry_policy=self.retry_policy,
            )
            debug_print(
                self.debug, workflow.now(), f"Retriever context: {retriever_result}"
            )
            context_prompt = f"\nHere is the context to use to answer the user's question:\n{retriever_result}"
            return context_prompt
        except ActivityError as e:
            workflow.logger.error(
                f"Failed to retrieve context from retriever. Context will not be added. Error: {e.cause}"
            )
            return ""

    async def handle_interrupt(self, tool_call: AgentToolCall):
        """
        Handles interrupts at a specified point ("before" or "after").
        """
        if tool_call.function.name in self.interrupt_map:
            self.interrupted.add(tool_call.id)
            interrupt = self.interrupt_map[tool_call.function.name]
            debug_print(
                self.debug,
                workflow.now(),
                f"Interrupt: {interrupt.question}",
            )
            self.orchestrator.reply(
                self.task_id,
                ResumeRequest(
                    tool_id=tool_call.id,
                    tool_name=tool_call.function.name,
                    question=interrupt.question,
                    when=interrupt.when,
                    tool_arguments=tool_call.function.arguments,
                    task_id=self.task_id,
                ),
            )

            await workflow.wait_condition(lambda: tool_call.id in self.resumed)

            debug_print(
                self.debug,
                workflow.now(),
                f"Interrupt: Resuming tool call '{tool_call.function.name}'",
            )

            resume_params = self.resumed[tool_call.id]

            if resume_params.action == "reject":
                raise ToolRejectedError("Tool rejected.") from ValueError(
                    f"Rejected by user. Reason: '{resume_params.content}'"
                )
            del self.resumed[tool_call.id]

    async def handle_tool_call(
        self,
        tool_call: AgentToolCall,
        context_variables: ContextVariables,
    ) -> AgentWorkflowResponse:
        """
        Handles a tool call, checks for interrupts before and after execution,
        and handles approval/rejection.
        """
        name = tool_call.function.name
        args = json.loads(tool_call.function.arguments)
        try:
            await self.handle_interrupt(tool_call)

            # Execute function in activity
            debug_print(
                self.debug,
                workflow.now(),
                f"Processing tool call: '{name}' with args: {args}",
            )
            result: AgentExecuteFnResult = await workflow.execute_activity(
                f"{self.agent.type}_execute_function",
                ExecuteFunctionParams(name, args, context_variables),
                result_type=AgentExecuteFnResult,
                start_to_close_timeout=self.start_to_close_timeout,
                retry_policy=self.retry_policy,
            )

            tool_response = ToolResponse(
                tool_call_id=tool_call.id,
                output=result,
            )
            return AgentWorkflowResponse(output=tool_response, sender=self.agent.name)
        except (ActivityError, ToolRejectedError) as e:
            # If error, let the model know by sending error message in tool response
            workflow.logger.error(
                f"Failed to process tool call '{name}'. "
                f"Error will be sent to agent to reassess. Error: {e}"
            )
            result = AgentExecuteFnResult(
                output=str(e.__cause__), context_variables=context_variables
            )
            tool_response = ToolResponse(tool_call_id=tool_call.id, output=result)
            return AgentWorkflowResponse(output=tool_response, sender=self.agent.name)



from .helpers import (
    create_retry_policy,
    debug_print,
    function_to_json,
    function_to_json_anthropic,
    mcp_to_anthropic_tool,
    mcp_to_openai_tool,
)

__all__ = [
    "create_retry_policy",
    "debug_print",
    "function_to_json",
    "function_to_json_anthropic",
    "mcp_to_anthropic_tool",
    "mcp_to_openai_tool",
]



from collections import deque
from dataclasses import dataclass
from typing import AsyncIterator, Literal, overload
from temporalio.client import (
    Client,
    Schedule,
    ScheduleActionStartWorkflow,
    ScheduleSpec,
    ScheduleHandle,
    WithStartWorkflowOperation,
    WorkflowHandle,
)
from temporalio import workflow, common
from temporalio.worker import Worker
from rojak.types import MCPServerConfig, InitMcpResult

with workflow.unsafe.imports_passed_through():
    from mcp import Tool
    from rojak.retrievers import RetrieverActivities
    from rojak.agents import AgentActivities
    from rojak.mcp import MCPClient
    from rojak.workflows import (
        OrchestratorResponse,
        OrchestratorParams,
        OrchestratorWorkflow,
        GetConfigResponse,
        UpdateConfigParams,
        TaskParams,
    )
    from rojak.agents import ResumeRequest, ResumeResponse
    from uuid import uuid4


@dataclass
class RunResponse:
    id: str
    result: OrchestratorResponse | ResumeRequest
    task_id: str
    workflow_handle: WorkflowHandle


class Rojak:
    def __init__(self, client: Client, task_queue: str):
        self.client: Client = client
        self.task_queue: str = task_queue
        self.mcp_result: InitMcpResult | None = None

    async def _init_mcp(self, servers: dict[str, MCPServerConfig]) -> None:
        """Initialise MCP servers.

        Args:
            servers (dict[str, MCPServerConfig]): List of MCP servers.

        Returns:
            tuple[dict[str, MCPClient], dict[str, Tool], dict[str, str]]: Response as tuple.
        """
        mcp_clients: dict[str, MCPClient] = {}
        mcp_tools: dict[str, Tool] = {}
        tool_client_mapping: dict[str, str] = {}
        for server_name, config in servers.items():
            try:
                mcp_client = MCPClient()
                await mcp_client.connect_to_server(config)
                list_tools_result = await mcp_client.session.list_tools()
                mcp_clients[server_name] = mcp_client
                for tool in list_tools_result.tools:
                    mcp_tools[tool.name] = tool
                    tool_client_mapping[tool.name] = server_name
            except Exception as e:
                print(f"Unable to connect to MCP server. Skipping. Error: {e}")
        self.mcp_result = InitMcpResult(mcp_clients, mcp_tools, tool_client_mapping)
        if mcp_tools:
            print(f"MCP tools loaded: {list(mcp_tools.keys())}")

    async def cleanup_mcp(self):
        """Cleanup MCP connections."""
        for client in list(self.mcp_result.clients.values())[::-1]:
            await client.cleanup()

    async def create_worker(
        self,
        agent_activities: list[AgentActivities],
        retriever_activities: list[RetrieverActivities] = [],
        mcp_servers: dict[str, MCPServerConfig] = {},
    ) -> Worker:
        """Create a worker.

        Args:
            agent_activities (list[AgentActivities]): List of activity classes that can be called.
            retriever_activities (list[RetrieverActivities], optional): List of retriever activity classes that can be called. Defaults to [].
            mcp_servers (dict[str, MCPServerConfig], optional): Dictionary of MCP server configurations. Each key represents the server name, and the value is its corresponding MCPServerConfig object. Defaults to {}.

        Returns:
            Worker: A worker object that can be used to start the worker.
        """
        await self._init_mcp(mcp_servers)
        activities = []
        for activity in agent_activities:
            if self.mcp_result:
                activity._add_mcp_configs(self.mcp_result)
            activities.append(activity.call)
            activities.append(activity.execute_function)
            activities.append(activity.execute_instructions)

        for retriever in retriever_activities:
            activities.append(retriever.retrieve_and_combine_results)

        worker: Worker = Worker(
            self.client,
            task_queue=self.task_queue,
            workflows=[OrchestratorWorkflow],
            activities=activities,
        )

        return worker

    async def list_scheduled_runs(
        self,
        schedule_id: str,
        statuses: list[
            Literal[
                "Running", "Completed", "Failed", "Cancelled", "Terminated", "TimedOut"
            ]
        ]
        | None = None,
        limit: int = 10,
        page_size: int = 1000,
        next_page_token: bytes | None = None,
    ) -> AsyncIterator[str]:
        """List the ID of orchestrators associated with the schedule.

        Args:
            schedule_id (str): Unique identifier of the schedule.
            statuses (list[ Literal[ 'Running', 'Completed', 'Failed', 'Cancelled', 'Terminated', 'TimedOut' ] ] | None, optional): List of statuses to filter the runs. Defaults to None.
            limit (int, optional): Maximum number of IDs to return. Defaults to 10.
            page_size (int, optional): Maximum number of results per page. Defaults to 1000.
            next_page_token (bytes | None, optional): A previously obtained next page token if doing pagination. Usually not needed as the iterator automatically starts from the beginning. Defaults to None.

        Returns:
            AsyncIterator[str]:  An async iterator that can be used with `async for`.
        """
        status_filter = (
            " OR ".join(f'ExecutionStatus="{status}"' for status in statuses)
            if statuses
            else ""
        )
        query = f'TemporalScheduledById="{schedule_id}"'
        if status_filter:
            query += f" AND ({status_filter})"

        async for workflow_execution in self.client.list_workflows(
            query=query,
            limit=limit,
            page_size=page_size,
            next_page_token=next_page_token,
        ):
            yield workflow_execution.id

    async def create_schedule(
        self,
        schedule_id: str,
        schedule_spec: ScheduleSpec,
        task: TaskParams,
        context_variables: dict = {},
        max_turns: int = float("inf"),
        history_size: int = 10,
        debug: bool = False,
    ) -> ScheduleHandle:
        """
        Create a schedule that periodically executes a workflow.

        The schedule periodically executes the equivalent of the `run()` method with the provided task, context variables, and configuration.

        Args:
            schedule_id (str): Unique identifier for the schedule.
            schedule_spec (ScheduleSpec): Specifies when the schedule executes, such as a cron schedule or interval.
            task (TaskParams): Encapsulates the agent, messages, and parameters for the workflow to run.
            context_variables (dict, optional): Additional variables available to functions and agent instructions. Defaults to {}.
            max_turns (int, optional): The maximum number of conversational turns allowed in the workflow. Defaults to float("inf").
            history_size (int, optional): The maximum number of messages retained in the conversation history. Defaults to 10.
            debug (bool, optional): Enables debug logging if True. Defaults to False.

        Returns:
            ScheduleHandle: A handle to the created schedule, allowing management such as pausing, resuming, or deleting.
        """
        task_id = str(uuid4())
        data = OrchestratorParams(
            context_variables=context_variables,
            max_turns=max_turns,
            tasks=deque([(task_id, task)]),
            debug=debug,
            type="stateless",
            history_size=history_size,
        )

        return await self.client.create_schedule(
            schedule_id,
            Schedule(
                action=ScheduleActionStartWorkflow(
                    OrchestratorWorkflow.run,
                    data,
                    id=schedule_id,
                    task_queue=self.task_queue,
                ),
                spec=schedule_spec,
            ),
        )

    @overload
    async def run(
        self,
        id: str,
        type: Literal["stateless", "persistent"],
        task: TaskParams,
        context_variables: dict = {},
        max_turns: int = float("inf"),
        history_size: int = 10,
        debug: bool = False,
    ) -> RunResponse: ...

    @overload
    async def run(self, id: str, resume: ResumeResponse) -> RunResponse: ...

    async def run(
        self,
        id: str,
        type: Literal["stateless", "persistent"] | None = None,
        task: TaskParams | None = None,
        resume: ResumeResponse | None = None,
        context_variables: dict = {},
        max_turns: int = float("inf"),
        history_size: int = 10,
        debug: bool = False,
    ) -> RunResponse:
        """
        Initialize and execute an orchestrator with the provided inputs, handling tasks, resuming workflows,
        and waiting for completion.

        Requires a running worker.

        Args:
            id (str): Unique identifier of the orchestrator.
            type (Literal["stateless", "persistent"] | None, optional): Whether to keep track of prior conversations through long-running workflows.
            task (TaskParams | None, optional): A task to be executed in the orchestrator. Defaults to None.
            resume (ResumeResponse | None, optional): A resume object for continuing a paused workflow. Defaults to None.
            context_variables (dict, optional): A dictionary of additional context variables available to functions
                and agent instructions. Defaults to an empty dictionary.
            max_turns (int, optional): The maximum number of conversational turns allowed. Defaults to infinity.
            history_size (int, optional): The maximum number of messages retained in the list before older messages are
                removed. When this limit is exceeded, older messages are removed. Defaults to 10.
            debug (bool, optional): If True, enables debug logging for the orchestrator. Defaults to False.

        Returns:
            RunResponse:
                - OrchestratorResponse: A response object containing updated messages, context variables, and agent
                  information, if the workflow completes successfully.
                - ResumeRequest: A request to resume the workflow if the current state requires further inputs or actions.
                - WorkflowHandle: A handle to the orchestrator workflow.

        Notes:
            - If a `task` is provided, the method starts a new orchestrator workflow.
            - If a `resume` is provided, the method resumes the specified workflow.
            - A new workflow is initialized with `context_variables`, `max_turns`, and `debug` settings.
        """
        start_op = None
        task_id = str(uuid4())
        if task:
            start_op = WithStartWorkflowOperation(
                OrchestratorWorkflow.run,
                OrchestratorParams(
                    context_variables=context_variables,
                    max_turns=max_turns,
                    debug=debug,
                    history_size=history_size,
                    type=type,
                ),
                id=id,
                id_conflict_policy=common.WorkflowIDConflictPolicy.USE_EXISTING,
                task_queue=self.task_queue,
            )
            result = await self.client.execute_update_with_start_workflow(
                OrchestratorWorkflow.add_task,
                (task_id, task),
                start_workflow_operation=start_op,
                result_type=OrchestratorResponse | ResumeRequest,
            )
            workflow_handle = await start_op.workflow_handle()
        else:
            workflow_handle = self.client.get_workflow_handle_for(
                workflow=OrchestratorWorkflow, workflow_id=id
            )

            result: (
                OrchestratorResponse | ResumeRequest
            ) = await workflow_handle.execute_update(
                OrchestratorWorkflow.add_task,
                (task_id, resume),
                result_type=OrchestratorResponse | ResumeRequest,
            )

        return RunResponse(
            id=workflow_handle.id,
            result=result,
            task_id=task_id,
            workflow_handle=workflow_handle,
        )

    async def get_result(
        self, id: str, task_id: str | None
    ) -> OrchestratorResponse | ResumeRequest | None:
        """
        Retrieve the latest or specific task result for a workflow.

        Requires a running worker. If `task_id` is provided, the result of that specific task
        is fetched; otherwise, the latest result of the workflow is returned.

        Args:
            id (str): The unique identifier of the workflow.
            task_id (str | None): The ID of the specific task to retrieve the result for. If None, retrieves the latest result.

        Returns:
            OrchestratorResponse: An object containing updated messages and context variables from the workflow.
        """
        workflow_handle = self.client.get_workflow_handle(id)
        if task_id is None:
            return await workflow_handle.query(OrchestratorWorkflow.get_latest_result)
        else:
            return await workflow_handle.query(OrchestratorWorkflow.get_result, task_id)

    async def get_config(self, id: str) -> GetConfigResponse:
        """
        Retrieve the current configuration of a workflow session.

        Requires a running worker.

        Args:
            id (str): The unique identifier of the workflow session.

        Returns:
            GetConfigResponse: An object containing the current configuration values of the session.
        """
        return await self.client.get_workflow_handle(id).query(
            OrchestratorWorkflow.get_config, result_type=GetConfigResponse
        )

    async def update_config(self, id: str, params: UpdateConfigParams):
        """
        Update the configuration of a workflow session.

        Requires a running worker.

        Args:
            id (str): The unique identifier of the workflow session.
            params (UpdateConfigParams): Configuration parameters to update. Only the values specified in `params` will be updated.
        """
        await self.client.get_workflow_handle(id).signal(
            OrchestratorWorkflow.update_config, params
        )

    async def cancel(self, id: str):
        """Cancel the session."""
        return await self.client.get_workflow_handle(id).cancel()



from dataclasses import dataclass
from typing import Any, Literal, Sequence
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rojak.mcp.mcp_client import MCPClient
    from mcp import Tool


@dataclass
class ConversationMessage:
    role: Literal["user", "assistant", "tool", "system"]
    """The role of the messages author."""

    content: str | None = None
    """The contents of the assistant message. Required unless tool_calls is specified."""

    tool_calls: list[Any] | None = None
    """The tool calls generated by the model, such as function calls."""

    tool_call_id: str | None = None
    """Unique identifier of the tool call."""

    sender: str | None = None
    """Indicate which agent the message originated from."""


ContextVariables = dict[str, Any]


@dataclass
class RetryPolicy:
    """Options for retrying agent activities."""

    initial_interval_in_seconds: int = 1
    """Backoff interval for the first retry. Default 1s."""

    backoff_coefficient: float = 2.0
    """Coefficient to multiply previous backoff interval by to get new
    interval. Default 2.0.
    """

    maximum_interval_in_seconds: int | None = None
    """Maximum backoff interval between retries. Default 100x
    :py:attr:`initial_interval`.
    """

    maximum_attempts: int = 0
    """Maximum number of attempts.
    
    If 0, the default, there is no maximum.
    """

    non_retryable_error_types: Sequence[str] | None = None
    """List of error types that are not retryable."""

    def __post_init__(self):
        # Validation taken from Go SDK's test suite
        if self.maximum_attempts == 1:
            # Ignore other validation if disabling retries
            return
        if self.initial_interval_in_seconds < 0:
            raise ValueError("Initial interval cannot be negative")
        if self.backoff_coefficient < 1:
            raise ValueError("Backoff coefficient cannot be less than 1")
        if self.maximum_interval_in_seconds:
            if self.maximum_interval_in_seconds < 0:
                raise ValueError("Maximum interval cannot be negative")
            if self.maximum_interval_in_seconds < self.initial_interval_in_seconds:
                raise ValueError(
                    "Maximum interval cannot be less than initial interval"
                )
        if self.maximum_attempts < 0:
            raise ValueError("Maximum attempts cannot be negative")


@dataclass
class RetryOptions:
    timeout_in_seconds: int = 60
    """Maximum time allowed for an agent to complete its tasks."""

    retry_policy: RetryPolicy | None = None
    """Options for retrying agent activities."""

    def __post_init__(self):
        if self.timeout_in_seconds < 1:
            raise ValueError("Timeout cannot be less than one second")


@dataclass
class MCPServerConfig:
    """Configuration options for the MCP server"""

    type: Literal["sse", "stdio"]
    """Connection type to MCP server."""

    command: str | None = None
    """(For `stdio` type) The command or executable to run to start the MCP server."""

    args: list[str] | None = None
    """(For `stdio` type) Command line arguments to pass to the `command`."""

    url: str | None = None
    """(For `websocket` or `sse` type) The URL to connect to the MCP server."""


@dataclass
class InitMcpResult:
    """Result from initialising MCP servers"""

    clients: dict[str, "MCPClient"]
    """A dictionary mapping server names to their corresponding `MCPClient` instances."""

    tools: dict[str, "Tool"]
    """A dictionary mapping tool names to `Tool` instances registered with the MCP servers."""

    tool_client_mapping: dict[str, str]
    """A dictionary mapping tool names to the corresponding MCP server names."""



from temporalio.client import ScheduleSpec, ScheduleIntervalSpec, ScheduleCalendarSpec
from .client import Rojak

__all__ = [
    "ScheduleSpec",
    "ScheduleIntervalSpec",
    "ScheduleCalendarSpec",
    "Rojak",
]



from .types import (
    ContextVariables,
    ConversationMessage,
    RetryPolicy,
    RetryOptions,
    MCPServerConfig,
    InitMcpResult,
)

__all__ = [
    "ContextVariables",
    "ConversationMessage",
    "RetryPolicy",
    "RetryOptions",
    "MCPServerConfig",
    "InitMcpResult",
]



from dataclasses import dataclass, field
import json
import os
from typing import Literal
from temporalio import activity
from anthropic import Anthropic, AnthropicBedrock, NotGiven
from anthropic.types import (
    Message,
    TextBlock,
    ToolUseBlock,
    MessageParam,
    ToolResultBlockParam,
    TextBlockParam,
)
from rojak.agents.agent import (
    Agent,
    AgentActivities,
    AgentCallParams,
    AgentExecuteFnResult,
    AgentOptions,
    AgentResponse,
    AgentToolCall,
    ExecuteFunctionParams,
    ExecuteInstructionsParams,
    ToolCallFunction,
)
from rojak.types.types import ConversationMessage
from rojak.utils import function_to_json_anthropic, mcp_to_anthropic_tool


@dataclass
class AnthropicAgentOptions(AgentOptions):
    api_key: str | None = None
    client: Anthropic | AnthropicBedrock | None = None
    inference_config: dict[str, any] = field(
        default_factory=lambda: {
            "max_tokens": 1000,
            "temperature": 0.0,
            "top_p": 0.9,
            "stop_sequences": [],
        }
    )


@dataclass
class AnthropicAgent(Agent):
    model: str = "claude-3-5-haiku-20241022"

    type: Literal["anthropic"] = field(default="anthropic")
    """Type of agent. Must be `"anthropic"`"""

    inference_config: dict[str, any] | None = None
    """Inference configuration for Anthropic models"""


class AnthropicAgentActivities(AgentActivities):
    def __init__(self, options: AnthropicAgentOptions = AnthropicAgentOptions()):
        super().__init__(options)

        if options.client:
            self.client = options.client
        elif options.api_key:
            self.client = Anthropic(api_key=options.api_key)
        elif os.environ.get("ANTHROPIC_API_KEY"):
            self.client = Anthropic()
        else:
            raise ValueError("Anthropic API key is required")

        self.inference_config = options.inference_config

    @staticmethod
    def handle_model_response(response: Message) -> AgentResponse:
        """Convert model response to AgentResponse"""
        content = ""
        tool_calls = []
        for block in response.content:
            if isinstance(block, TextBlock):
                content += block.text
            elif isinstance(block, ToolUseBlock):
                tool_calls.append(
                    AgentToolCall(
                        id=block.id,
                        function=ToolCallFunction(
                            name=block.name,
                            arguments=json.dumps(block.input),
                        ),
                    )
                )

        if tool_calls:
            return AgentResponse(type="tool", content=content, tool_calls=tool_calls)
        else:
            return AgentResponse(type="text", tool_calls=tool_calls, content=content)

    @staticmethod
    def convert_messages(
        messages: list[ConversationMessage],
    ) -> tuple[list[MessageParam], str | None]:
        """Convert messages to be Anthropic compatible.

        1. System messages become the returned system string (only one).
        2. User/assistant text messages remain single messages.
        3. Assistant messages that contain tool calls become a single message
        with a list of tool-use blocks.
        4. Consecutive tool-result messages are combined into one user message,
        where each result is a separate 'tool_result' block.
        """
        converted_messages: list[MessageParam] = []
        system_message: str | None = None

        # Temporary storage for consecutive tool-result blocks
        tool_result_buffer: list[ToolResultBlockParam] = []

        def flush_tool_results():
            # If we have accumulated any tool results,
            # append them as a single user message and clear the buffer.
            nonlocal tool_result_buffer
            if tool_result_buffer:
                converted_messages.append(
                    MessageParam(role="user", content=tool_result_buffer)
                )
                tool_result_buffer = []

        for msg in messages:
            if msg.role == "system":
                system_message = msg.content

            elif msg.tool_calls:
                # Whenever we hit an assistant message with tool calls,
                # first flush any pending tool results (belonging to 'tool' role messages).
                flush_tool_results()

                # Convert each tool call into a tool-use block.
                tool_blocks = []
                for tool_call_dict in msg.tool_calls:
                    tool_call = AgentToolCall(**tool_call_dict)
                    tool_blocks.append(
                        ToolUseBlock(
                            type="tool_use",
                            id=tool_call.id,
                            input=json.loads(tool_call.function.arguments),
                            name=tool_call.function.name,
                        ).model_dump()
                    )

                # Append as a single assistant message containing multiple tool calls.
                converted_messages.append(
                    MessageParam(role="assistant", content=tool_blocks)
                )

            elif msg.role == "tool":
                # We accumulate tool results into a buffer until we reach a non-tool message.
                tool_result_buffer.append(
                    ToolResultBlockParam(
                        type="tool_result",
                        tool_use_id=msg.tool_call_id,
                        content=[TextBlockParam(type="text", text=msg.content)],
                    )
                )

            else:
                # For normal user/assistant messages, first flush any pending tool results.
                flush_tool_results()

                # Then add this user/assistant message as-is.
                if msg.role in ["user", "assistant"]:
                    converted_messages.append(
                        MessageParam(role=msg.role, content=msg.content)
                    )

        # If conversation ended with tool results, flush them.
        flush_tool_results()

        return converted_messages, system_message

    @activity.defn(name="anthropic_call")
    async def call(self, params: AgentCallParams) -> AgentResponse:
        # Create list of messages
        messages, system_message = self.convert_messages(params.messages)

        if params.inference_config:
            self.inference_config = {**self.inference_config, **params.inference_config}

        # Create tool call json
        functions = [
            self.function_map[name]
            for name in params.function_names
            if name in self.function_map
        ]
        tools = [function_to_json_anthropic(f) for f in functions]
        for tool in tools:
            fn_params = tool["input_schema"]["properties"]
            fn_params.pop("context_variables", None)
            required_list = tool["input_schema"]["required"]
            if "context_variables" in required_list:
                required_list.remove("context_variables")

        tools += [
            mcp_to_anthropic_tool(tool) for tool in self.mcp_result.tools.values()
        ]

        response: Message = self.client.messages.create(
            model=params.model,
            messages=messages,
            system=system_message or NotGiven(),
            tools=tools or NotGiven(),
            tool_choice=params.tool_choice
            if tools and params.tool_choice
            else NotGiven(),
            max_tokens=self.inference_config["max_tokens"],
            temperature=self.inference_config["temperature"],
            top_p=self.inference_config["top_p"],
            stop_sequences=self.inference_config["stop_sequences"],
        )

        return self.handle_model_response(response)

    @activity.defn(name="anthropic_execute_instructions")
    async def execute_instructions(self, params: ExecuteInstructionsParams) -> str:
        return await super().execute_instructions(params)

    @activity.defn(name="anthropic_execute_function")
    async def execute_function(
        self, params: ExecuteFunctionParams
    ) -> str | AnthropicAgent | AgentExecuteFnResult:
        return await super().execute_function(params)



from .retriever import Retriever, RetrieverActivities

try:
    from .qdrant_retriever import (  # noqa: F401
        QdrantRetriever,
        QdrantRetrieverActivities,
        QdrantRetrieverOptions,
    )

    _QDRANT_AVAILABLE_ = True
except ImportError:
    _QDRANT_AVAILABLE_ = False

__all__ = ["Retriever", "RetrieverActivities"]


if _QDRANT_AVAILABLE_:
    __all__.extend(
        ["QdrantRetriever", "QdrantRetrieverActivities", "QdrantRetrieverOptions"]
    )



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Callable, Literal
from mcp.types import TextContent
from rojak.retrievers import Retriever
from rojak.types import (
    ContextVariables,
    ConversationMessage,
    RetryOptions,
    MCPServerConfig,
    InitMcpResult,
)
from temporalio.exceptions import ApplicationError

AgentFunction = Callable[[], str]


@dataclass
class AgentOptions:
    all_functions: list[AgentFunction] = field(default_factory=list)
    """List of functions that an agent can execute."""

    mcp_servers: dict[str, MCPServerConfig] = field(default_factory=dict)
    """List of MCP servers to connect to."""


@dataclass
class AgentCallParams:
    """Parameters for generating response from model."""

    messages: list[ConversationMessage]
    """List of message objects."""

    model: str
    """The LLM model to use."""

    function_names: list[str] = field(default_factory=list)
    """List of tool call function names that the agent can select."""

    inference_config: dict[str, Any] | None = None
    """Additional configurations for the inference."""

    parallel_tool_calls: bool = True
    """Whether model should perform multiple tool calls together."""

    tool_choice: Any | None = None
    """The tool choice for the agent, if any."""


@dataclass
class ExecuteFunctionParams:
    """Parameters for executing tool call function."""

    name: str
    """The name of the tool call function."""

    args: dict[str, Any]
    """The arguments for the tool call function."""

    context_variables: ContextVariables
    """A dictionary of additional context variables, available to functions and Agent instructions."""


@dataclass
class ExecuteInstructionsParams:
    instructions: "AgentInstructionOptions"
    """Options for the callable instructions."""

    context_variables: ContextVariables
    """A dictionary of additional context variables, available to functions and Agent instructions."""


@dataclass
class ToolCallFunction:
    arguments: str
    """String representation of the arguments for the tool call function."""

    name: str
    """The name of the tool call function."""


@dataclass
class AgentToolCall:
    id: str
    """Unique identifier of the tool call."""

    function: ToolCallFunction
    """Function that the model called."""

    type: str = "function"
    """The type of the tool."""

    index: int | None = None
    """Identifies which function call the delta is for."""

    def __post_init__(self):
        if isinstance(self.function, dict):
            self.function = ToolCallFunction(**self.function)


@dataclass
class AgentResponse:
    """Response object from generating response from model."""

    type: Literal["text", "tool"]
    """Specify if it is a natural language or a tool call response."""

    content: str | None = None
    """String output from the model"""

    tool_calls: list[AgentToolCall] | None = None
    """List of tool call objects."""

    def __post_init__(self):
        if self.tool_calls:
            self.tool_calls = [
                tool_call
                if isinstance(tool_call, AgentToolCall)
                else AgentToolCall(**tool_call)
                for tool_call in self.tool_calls
            ]


@dataclass
class AgentInstructionOptions:
    """Information of the callable instructions."""

    type: Literal["function"]
    """The type of the instruction. Only `function` is supported."""

    name: str
    """The name of the function."""


@dataclass
class Interrupt:
    tool_name: str
    """The name of the tool to interrupt."""

    question: str = ""
    """The question to ask the user."""

    when: Literal["before"] = "before"
    """When the interrupt should be triggered."""


@dataclass
class ResumeRequest:
    """Request to resume the interrupted agent."""

    tool_id: str
    """The ID of the tool that is interrupted."""

    tool_arguments: str
    """Arguments that will be passed to the tool that was interrupted."""

    task_id: str
    """Unique identifier of the request that triggered the interrupt."""

    tool_name: str
    """The name of the tool to interrupt."""

    question: str = ""
    """The question to ask the user."""

    when: Literal["before"] = "before"
    """When the interrupt should be triggered."""


@dataclass
class ResumeResponse:
    """Response to resume the interrupted agent."""

    action: Literal["approve", "reject"]
    """Action to take on the interrupt."""

    tool_id: str
    """Tool call id to resume."""

    content: str | None = None
    """Feedback to pass to Agent. Only for 'rejected' action."""


@dataclass
class Agent(ABC):
    model: str
    """The LLM model to use."""

    type: str
    """The prefix of the activity name."""

    name: str = "Agent"
    """The name of the agent."""

    instructions: str | AgentInstructionOptions = "You are a helpful assistant."
    """Instructions for the agent, can be a string or a callable returning a string."""

    functions: list[str] = field(default_factory=list)
    """A list of functions that the agent can call."""

    tool_choice: Any | None = None
    """The tool choice for the agent, if any."""

    parallel_tool_calls: bool = True
    """Whether model should perform multiple tool calls together."""

    interrupts: list[Interrupt] = field(default_factory=list)
    """List of interrupts for reviewing tool use."""

    retriever: Retriever | None = None
    """Specify which retriever to use."""

    retry_options: RetryOptions = field(default_factory=RetryOptions)
    """Options for timeout and retries."""


@dataclass
class AgentExecuteFnResult:
    """Result object from executing tool call function."""

    output: str = ""
    """String output to pass as message content."""

    agent: Agent | None = None
    """The agent to call next."""

    context_variables: ContextVariables = field(default_factory=dict)
    """A dictionary of additional context variables, available to functions and Agent instructions."""


class AgentActivities(ABC):
    """
    Abstract base class for Agent implementations.
    This class provides a common structure for different types of agents.
    """

    def __init__(self, options: AgentOptions):
        self.function_map = {f.__name__: f for f in options.all_functions}
        self.mcp_result: InitMcpResult | None = None

    def _add_mcp_configs(self, mcp_result: InitMcpResult):
        """Add MCP configurations"""
        self.mcp_result = mcp_result

    @abstractmethod
    async def call(self, params: AgentCallParams) -> AgentResponse:
        """Generate response from the LLM model.

        Args:
            params (AgentCallParams): Parameters for response generation.

        Returns:
            AgentResponse: Generated response from the model.
        """
        pass

    @abstractmethod
    async def execute_instructions(self, params: ExecuteInstructionsParams) -> str:
        """Execute the instruction callable.

        Args:
            params (ExecuteInstructionsParams): Parameters containing information for executing callable.

        Raises:
            ApplicationError: Error occurred while executing instructions.

        Returns:
            str: Instructions as a string.
        """
        instructions = params.instructions
        if instructions.name not in self.function_map:
            raise ApplicationError(
                f"Function {instructions.name} not found",
                type="FunctionNotFound",
                non_retryable=True,
            )

        fn = self.function_map[instructions.name]
        args = {}

        if "context_variables" in fn.__code__.co_varnames:
            args["context_variables"] = params.context_variables

        res = fn(**args)
        return str(res)

    def handle_function_result(
        self,
        result: str | Agent | AgentExecuteFnResult,
        context_variables: ContextVariables,
    ) -> AgentExecuteFnResult:
        match result:
            case str():
                return AgentExecuteFnResult(
                    output=result,
                    context_variables=context_variables,
                )
            case Agent():
                return AgentExecuteFnResult(
                    output=f"Transferred to '{result.name}'",
                    agent=result,
                    context_variables=context_variables,
                )
            case AgentExecuteFnResult():
                return result
            case _:
                try:
                    return AgentExecuteFnResult(
                        output=str(result), context_variables=context_variables
                    )
                except Exception as e:
                    raise TypeError(
                        f"Unknown function result type: {type(result)}. Error: {str(e)}"
                    )

    @staticmethod
    async def execute_mcp_tool(
        mcp_result: InitMcpResult, tool_name: str, args: dict
    ) -> str:
        """Get tool response from MCP server

        Args:
            mcp_result (InitMcpResult): The result from initialising MCP servers.
            tool_name (str): Name of the tool.
            args (dict): Tool arguments.

        Returns:
            str: The tool response.
        """
        server_name = mcp_result.tool_client_mapping[tool_name]
        client = mcp_result.clients[server_name]
        response = await client.session.call_tool(tool_name, args)
        texts = []
        for content in response.content:
            if isinstance(content, TextContent):
                texts.append(content.text)
        return "\n".join(texts)

    @abstractmethod
    async def execute_function(
        self, params: ExecuteFunctionParams
    ) -> str | Agent | AgentExecuteFnResult:
        """Execute the tool call function

        Args:
            params (ExecuteFunctionParams): Parameters for executing tool call function.

        Raises:
            ApplicationError: Error executing tool call function.

        Returns:
            str | Agent | AgentExecuteFnResult: Response from the tool call function.
        """
        if params.name in self.function_map:
            fn = self.function_map[params.name]

            if "context_variables" in fn.__code__.co_varnames:
                params.args["context_variables"] = params.context_variables

            result = fn(**params.args)
        elif self.mcp_result and params.name in self.mcp_result.tools:
            result = await self.execute_mcp_tool(
                self.mcp_result, params.name, params.args
            )
        else:
            raise ApplicationError(
                f"Function {params.name} not found",
                type="FunctionNotFound",
                non_retryable=True,
            )

        return self.handle_function_result(result, params.context_variables)



import copy
from dataclasses import dataclass, field
from typing import Literal
from temporalio import workflow
from temporalio.exceptions import ActivityError
from rojak.types import ConversationMessage, ContextVariables
from collections import deque
import asyncio
from rojak.utils import debug_print
from rojak.workflows.agent_workflow import (
    AgentWorkflow,
    AgentWorkflowRunParams,
    ResumeRequest,
    ResumeResponse,
    ToolResponse,
    AgentTypes,
)
from rojak.agents import Agent, Interrupt


@dataclass
class OrchestratorParams:
    type: Literal["stateless", "persistent"]
    """Specify if it is stateless or persistent workflow."""

    context_variables: ContextVariables = field(default_factory=dict)
    """A dictionary of additional context variables, available to functions and Agent instructions."""

    max_turns: int | float = field(default=float("inf"))
    """The maximum number of conversational turns allowed."""

    debug: bool = False
    """If True, enables debug logging"""

    history_size: int = field(default=10)
    """The maximum number of messages retained in the list before older messages are removed."""

    messages: list[ConversationMessage] = field(default_factory=list)
    """List of conversation messages to initialise workflow with."""

    tasks: deque[tuple[str, "TaskParams"]] = field(default_factory=deque)
    """Tasks queue to initialise workflow with."""


@dataclass
class OrchestratorResponse:
    """The response object from containing the updated state."""

    messages: list[ConversationMessage]
    """The list of updated messages."""

    context_variables: ContextVariables
    """The dictionary of the updated context variables."""

    agent: AgentTypes | None = None
    """The last agent to be called."""

    interrupt: Interrupt | None = None
    """The object surfaced to the client when the interupt is triggered."""


@dataclass
class TaskParams:
    messages: list[ConversationMessage]
    """List of message object."""

    agent: AgentTypes
    """The agent to be called."""


@dataclass
class UpdateConfigParams:
    messages: list[ConversationMessage] | None = None
    """A list of message objects."""

    context_variables: ContextVariables | None = None
    """The dictionary of the updated context variables."""

    max_turns: int | float | None = None
    """The maximum number of conversational turns allowed."""

    history_size: int | None = None
    """The maximum number of messages retained in the list before older messages are removed."""

    debug: bool | None = None
    """If True, enables debug logging"""


@dataclass
class GetConfigResponse:
    messages: list[ConversationMessage]
    """A list of message objects."""

    context_variables: ContextVariables
    """The dictionary of the updated context variables."""

    max_turns: int | float
    """The maximum number of conversational turns allowed."""

    history_size: int
    """The maximum number of messages retained in the list before older messages are removed."""

    debug: bool
    """If True, enables debug logging"""


@workflow.defn
class OrchestratorWorkflow:
    @workflow.init
    def __init__(self, params: OrchestratorParams) -> None:
        self.lock = asyncio.Lock()  # Prevent concurrent update handler executions
        self.tasks: deque[tuple[str, TaskParams]] = params.tasks
        self.responses: dict[str, OrchestratorResponse | ResumeRequest] = {}
        self.latest_response: OrchestratorResponse | ResumeRequest | None = None
        self.max_turns = params.max_turns
        self.debug = params.debug
        self.context_variables = params.context_variables
        self.current_agent_workflow: AgentWorkflow | None = None
        self.task_id: str | None = None
        self.history_size = params.history_size
        self.type = params.type
        self.messages: list[ConversationMessage] = params.messages

    @workflow.run
    async def run(self, params: OrchestratorParams) -> OrchestratorResponse:
        while True:
            await workflow.wait_condition(lambda: bool(self.tasks))
            task_id, task = self.tasks.popleft()
            self.task_id = task_id
            self.messages += task.messages
            self.agent = task.agent  # Keep track of the last to be called

            message = self.messages[-1]
            debug_print(
                self.debug, workflow.now(), f"{message.role}: {message.content}"
            )

            active_agent = self.agent
            init_len = len(self.messages)
            past_message_state = copy.deepcopy(self.messages)

            try:
                while len(self.messages) - init_len < self.max_turns and active_agent:
                    active_agent = await self.process(active_agent)

                response = OrchestratorResponse(
                    messages=self.messages,
                    agent=self.agent,
                    context_variables=self.context_variables,
                )

                self.reply(self.task_id, response)

                await workflow.wait_condition(lambda: workflow.all_handlers_finished())

                if self.type == "stateless":
                    return self.responses[self.task_id]
                else:
                    if len(self.messages) > self.history_size:
                        messages = deque(self.messages[-self.history_size :])
                        while messages and messages[0].role == "tool":
                            messages.popleft()
                        self.messages = list(messages)

                    workflow_history_size = workflow.info().get_current_history_size()
                    workflow_history_length = (
                        workflow.info().get_current_history_length()
                    )
                    if (
                        workflow_history_length > 10_000
                        or workflow_history_size > 20_000_000
                    ):
                        debug_print(
                            self.debug,
                            workflow.now(),
                            "Continue as new due to prevent workflow event history from exceeding limit.",
                        )
                        workflow.continue_as_new(
                            args=[
                                OrchestratorParams(
                                    type=params.type,
                                    context_variables=self.context_variables,
                                    max_turns=self.max_turns,
                                    debug=self.debug,
                                    history_size=self.history_size,
                                    messages=self.messages,
                                    tasks=self.tasks,
                                )
                            ]
                        )
            except ActivityError as e:
                # Return messages to previous state and wait for new messages
                workflow.logger.error(f"Agent failed to complete. Error: {e}")
                self.messages = past_message_state
                active_agent = None
                self.pending = False
                continue

    def reply(self, task_id: str, response: OrchestratorResponse | ResumeRequest):
        """Return response back"""
        self.responses[task_id] = response
        self.latest_response = response

    async def process(self, active_agent: Agent) -> Agent | None:
        params = AgentWorkflowRunParams(
            agent=active_agent,
            messages=self.messages,
            context_variables=self.context_variables,
            debug=self.debug,
            orchestrator=self,
            task_id=self.task_id,
        )
        agent_workflow = AgentWorkflow(params)
        self.current_agent_workflow = agent_workflow
        response, updated_messages = await agent_workflow.run()

        self.messages = updated_messages

        if isinstance(response.output, ToolResponse):
            fn_result = response.output.output
            if fn_result.agent is not None:
                debug_print(
                    self.debug,
                    workflow.now(),
                    f"{active_agent.name}: Transferred to '{fn_result.agent.name}'.",
                )
                self.agent = active_agent = fn_result.agent
            if fn_result.context_variables is not None:
                self.context_variables = fn_result.context_variables
        elif isinstance(response.output, str):
            debug_print(
                self.debug,
                workflow.now(),
                f"\n{active_agent.name}: {response.output}",
            )
            active_agent = None

        return active_agent

    def resume(self, params: ResumeResponse):
        """Resumes an interrupted agent workflow for a specific tool ID."""
        if not self.current_agent_workflow:
            raise ValueError("Cannot resume: No active agent workflow available.")

        tool_id = params.tool_id
        if tool_id in self.current_agent_workflow.interrupted:
            self.current_agent_workflow.interrupted.remove(tool_id)
            self.current_agent_workflow.resumed[tool_id] = params
        else:
            raise KeyError(
                f"Cannot resume: Tool ID '{tool_id}' not found in the approval queue."
            )

    @workflow.query
    def get_messages(self) -> list[ConversationMessage]:
        return self.messages

    @workflow.update
    async def add_task(
        self,
        params: tuple[str, TaskParams | ResumeResponse],
    ) -> OrchestratorResponse | ResumeRequest:
        task_id, task = params
        async with self.lock:
            self.task_id = task_id
            if isinstance(task, TaskParams):
                self.tasks.append(params)
            else:
                self.resume(task)
            await workflow.wait_condition(lambda: task_id in self.responses)
            return self.responses[task_id]

    @workflow.query
    def get_result(self, task_id: str) -> OrchestratorResponse:
        return self.responses[task_id]

    @workflow.query
    def get_latest_result(self) -> OrchestratorResponse | ResumeRequest | None:
        return self.latest_response

    @workflow.query
    def get_config(self) -> GetConfigResponse:
        return GetConfigResponse(
            messages=self.messages,
            context_variables=self.context_variables,
            max_turns=self.max_turns,
            history_size=self.history_size,
            debug=self.debug,
        )

    @workflow.signal
    def update_config(self, params: UpdateConfigParams):
        if params.messages is not None:
            self.messages = params.messages
        if params.context_variables is not None:
            self.context_variables = params.context_variables
        if params.max_turns is not None:
            self.max_turns = params.max_turns
        if params.history_size is not None:
            self.history_size = params.history_size
        if params.debug is not None:
            self.debug = params.debug



from mcp import ClientSession, StdioServerParameters
from contextlib import AsyncExitStack
from mcp.client.stdio import stdio_client
from mcp.client.sse import sse_client
from rojak.types import MCPServerConfig


class MCPClient:
    def __init__(self):
        self.session: ClientSession | None = None
        self.exit_stack = AsyncExitStack()

    async def connect_to_server(self, config: MCPServerConfig):
        """Connect to an MCP server"""

        if config.type == "stdio":
            params = StdioServerParameters(command=config.command, args=config.args)
            stdio_transport = await self.exit_stack.enter_async_context(
                stdio_client(params)
            )
            self.stdio, self.write = stdio_transport
            self.session = await self.exit_stack.enter_async_context(
                ClientSession(self.stdio, self.write)
            )
        else:
            stdio_transport = await self.exit_stack.enter_async_context(
                sse_client(config.url)
            )
            self.sse, self.write = stdio_transport
            self.session = await self.exit_stack.enter_async_context(
                ClientSession(self.sse, self.write)
            )
        await self.session.initialize()

    async def cleanup(self):
        """Clean up resources"""
        await self.exit_stack.aclose()



from dataclasses import dataclass, field
from rojak.retrievers import RetrieverActivities, Retriever
from temporalio import activity
from qdrant_client import QdrantClient
from qdrant_client.qdrant_fastembed import QueryResponse


@dataclass
class QdrantRetrieverOptions:
    url: str
    collection_name: str


@dataclass
class QdrantRetriever(Retriever):
    type: str = field(init=False, default="qdrant")


class QdrantRetrieverActivities(RetrieverActivities):
    def __init__(self, options: QdrantRetrieverOptions):
        super().__init__(options)
        self.client = QdrantClient(url=options.url)
        self.collection_name = options.collection_name

    async def retrieve(self, text: str) -> list[QueryResponse]:
        retrieval_results = self.client.query(
            collection_name=self.collection_name, query_text=text
        )
        return retrieval_results

    @activity.defn(name="qdrant_retrieve_and_combine_results")
    async def retrieve_and_combine_results(self, text: str) -> str:
        retrieval_results = await self.retrieve(text)
        return self.combine_retreival_results(retrieval_results)

    @staticmethod
    def combine_retreival_results(results: list[QueryResponse]) -> str:
        return "\n".join([result.document for result in results])



from abc import ABC, abstractmethod
from dataclasses import dataclass


@dataclass
class Retriever(ABC):
    type: str
    """The prefix of the activity name."""


class RetrieverActivities(ABC):
    """
    Abstract base class for Retriever implementations.
    This class provides a common structure for different types of retrievers.
    """

    def __init__(self, options: any):
        self._options = options
        pass

    @abstractmethod
    async def retrieve(self, text: str) -> any:
        """Retrieve information based on the input text.

        This abstract method must be implemented by all concrete subclasses. It handles the retrieval of
        information relevant to the provided text.

        Args:
            text (str): The input text to base retrieval on.

        Returns:
            Any: The retrieved information corresponding to the input text.
        """
        pass

    @abstractmethod
    async def retrieve_and_combine_results(self, text: str) -> str:
        """Retrieve information and combine the results.

        This abstract method must be implemented by all concrete subclasses. It is responsible for performing
        data retrieval based on the input text and combining or processing the results into a final output.

        Args:
            text (str): The input text text to guide the retrieval process.

        Returns:
            str: The processed and combined results of the retrieval operation.
        """
        pass



from datetime import datetime, timedelta
import inspect
from rojak.types import RetryPolicy
from temporalio.common import RetryPolicy as TRetryPolicy
from mcp import Tool


def function_to_json(func) -> dict:
    """
    Converts a Python function into a JSON-serializable dictionary
    that describes the function's signature, including its name,
    description, and parameters.

    Compatible with OpenAI tool definition.

    Args:
        func: The function to be converted.

    Returns:
        A dictionary representing the function's signature in JSON format.
    """
    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
        type(None): "null",
    }

    try:
        signature = inspect.signature(func)
    except ValueError as e:
        raise ValueError(
            f"Failed to get signature for function {func.__name__}: {str(e)}"
        )

    parameters = {}
    for param in signature.parameters.values():
        try:
            param_type = type_map.get(param.annotation, "string")
        except KeyError as e:
            raise KeyError(
                f"Unknown type annotation {param.annotation} for parameter {param.name}: {str(e)}"
            )
        parameters[param.name] = {"type": param_type}

    required = [
        param.name
        for param in signature.parameters.values()
        if param.default == inspect._empty
    ]

    return {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": func.__doc__ or "",
            "parameters": {
                "type": "object",
                "properties": parameters,
                "required": required,
            },
        },
    }


def function_to_json_anthropic(func) -> dict:
    """
    Converts a Python function into a JSON-serializable dictionary
    that describes the function's signature, including its name,
    description, and parameters.

    Compatible with Anthropic tool definition.

    Args:
        func: The function to be converted.

    Returns:
        A dictionary representing the function's signature in JSON format.
    """
    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
        type(None): "null",
    }

    try:
        signature = inspect.signature(func)
    except ValueError as e:
        raise ValueError(
            f"Failed to get signature for function {func.__name__}: {str(e)}"
        )

    parameters = {}
    for param in signature.parameters.values():
        try:
            param_type = type_map.get(param.annotation, "string")
        except KeyError as e:
            raise KeyError(
                f"Unknown type annotation {param.annotation} for parameter {param.name}: {str(e)}"
            )
        parameters[param.name] = {"type": param_type}

    required = [
        param.name
        for param in signature.parameters.values()
        if param.default == inspect._empty
    ]

    return {
        "name": func.__name__,
        "description": func.__doc__ or "",
        "input_schema": {
            "type": "object",
            "properties": parameters,
            "required": required,
        },
    }


def mcp_to_openai_tool(tool: Tool) -> dict[str, any]:
    """Convert MCP tool to openai format.

    Args:
        tool (Tool): MCP Tool object.

    Returns:
        dict[str, any]: A dictionary representing tool in JSON format.
    """
    return {
        "type": "function",
        "function": {
            "name": tool.name,
            "description": tool.description,
            "parameters": tool.inputSchema,
        },
    }


def mcp_to_anthropic_tool(tool: Tool) -> dict[str, any]:
    """Convert MCP tool to anthropic format

    Args:
        tool (Tool): MCP Tool object

    Returns:
        dict[str, any]: A dictionary representing tool in JSON format.
    """
    return {
        "name": tool.name,
        "description": tool.description,
        "input_schema": tool.inputSchema,
    }


def create_retry_policy(retry_policy: RetryPolicy | None) -> TRetryPolicy | None:
    """Convert serialisable retry policy to Temporal RetryPolicy."""
    if retry_policy is None:
        return None

    initial_interval = (
        timedelta(seconds=retry_policy.initial_interval_in_seconds)
        if retry_policy.initial_interval_in_seconds
        else timedelta(seconds=1)
    )

    maximum_interval = (
        timedelta(seconds=retry_policy.maximum_interval_in_seconds)
        if retry_policy.maximum_interval_in_seconds
        else None
    )

    return TRetryPolicy(
        initial_interval=initial_interval,
        backoff_coefficient=retry_policy.backoff_coefficient,
        maximum_interval=maximum_interval,
        maximum_attempts=retry_policy.maximum_attempts,
        non_retryable_error_types=retry_policy.non_retryable_error_types,
    )


def debug_print(debug: bool, timestamp: datetime, *args: str) -> None:
    if not debug:
        return
    message = " ".join(map(str, args))
    print(f"\033[97m[\033[90m{timestamp}\033[97m]\033[90m {message}\033[0m")
